"""Exceptions used by Faros Client"""


class FarosException(Exception):
    """
    Base class for all faros client exceptions.
    All custom exceptions must be derived from this class
    """
    pass


class FarosGraphQLException(FarosException):
    """Raise when there is an error with the grapgql api such as on query"""
    pass



"""Faros API client"""

import json
from graphqlclient import GraphQLClient
from faros.exceptions import FarosGraphQLException

DEFAULT_API_URL = "https://api.faros.ai/v1"


class FarosClient:
    """Base client for Faros API"""

    def __init__(self, token, api_url=None):
        """
        Initializes the Faros Client. Using this directly is not encouraged.
        Users should generally use the class methods to initialize the client.

        :param str token: Faros token to authenticate with into the api
        :param str api_url: Faros service url to run api calls on

        :returns: Faros Client
        :rtype: FarosClient
        """

        self._api_url = api_url if api_url else DEFAULT_API_URL
        self._graphql_client = GraphQLClient(f"{self._api_url}/graphql")
        self._graphql_client.inject_token(f"Bearer {token}")

    @classmethod
    def from_event(cls, event, api_url=None):
        """
        Factory method to instantiate Faros Client from event object passed
        to the app handler

        :param dict event: object with properties
        :param str api_url: Faros service api url to query against

        :returns: instance of the Faros client
        :rtype: FarosClient
        """
        return cls(event["farosToken"], api_url)

    def graphql_query(self, query, variables=None):
        """
        Queries the Faros GraphQL api

        :param query: GraphQL query
        :param variables: graphql query variables
        :return: the GraphQL results
        :rtype: json
        """

        response = json.loads(self._graphql_client.execute(query, variables))
        if response.get("errors"):
            raise FarosGraphQLException(json.dumps(response["errors"]))
        return response["data"]



import unittest
from faros import utils
from datetime import datetime, timedelta


class UtilsTest(unittest.TestCase):
    def test_list_of_identical_scalars(self):
        list1 = ["a", "b", "c", "d"]
        diff = utils.diff_lists_of_scalars(list1, list1)
        self.assertFalse(diff.is_diff())

    def test_list_of_scalars_additions(self):
        list1 = ["a", "b", ]
        list2 = ["a", "b", "c", "d"]
        diff = utils.diff_lists_of_scalars(list1, list2)
        self.assertEqual(0, len(diff.removed_items()))
        self.assertEqual(2, len(diff.added_items()))

    def test_list_of_scalars_removals(self):
        list1 = ["a", "b", "c", "d"]
        list2 = ["a", ]
        diff = utils.diff_lists_of_scalars(list1, list2)
        self.assertEqual(3, len(diff.removed_items()))

    def test_list_of_scalars_mixed(self):
        list1 = [1, 2, 3, "c", "d"]
        list2 = [1, "c", "d", "e"]
        diff = utils.diff_lists_of_scalars(list1, list2)
        self.assertEqual(2, len(diff.removed_items()))
        self.assertEqual(1, len(diff.added_items()))
        self.assertEqual(["2", "3", "e"], [x.item for x in diff.diffs])

    def test_list_of_empty_scalars(self):
        list1 = []
        diff = utils.diff_lists_of_scalars(list1, list1)
        self.assertFalse(diff.is_diff())

    def test_list_of_empty_objects(self):
        list1 = utils.DiffObj("", "", [])
        diff = utils.diff_lists(list1, list1)
        self.assertFalse(diff.is_diff())

    def test_diff_lists_only_additions(self):
        list1 = utils.DiffObj("", "", [])
        list2 = utils.DiffObj("", "", [{}, {}, {}])
        diff = utils.diff_lists(list1, list2)
        self.assertEqual(3, len(diff.added_items()))

    def test_diff_lists_only_removals(self):
        list1 = utils.DiffObj("", "", [{}, {}])
        list2 = utils.DiffObj("", "", [])
        diff = utils.diff_lists(list1, list2)
        self.assertEqual(2, len(diff.removed_items()))

    def test_diff_lists_scalars(self):
        list1 = utils.DiffObj("", "", [1, 2, 3])
        list2 = utils.DiffObj("", "", [3, 4, 5])
        diff = utils.diff_lists(list1, list2)
        self.assertEqual(2, len(diff.removed_items()))
        self.assertEqual(2, len(diff.added_items()))

    def test_diff_empty_lists_of_objects(self):
        list1 = utils.DiffObj("", "", [])
        diff = utils.diff_lists_of_objects(list1, list1, {"": ""})
        self.assertFalse(diff.is_diff())

    def test_diff_lists_of_objects_only_additions(self):
        list1 = utils.DiffObj("", "", [])
        list2 = utils.DiffObj("", "", [{"key": 1}, {"key": 3}, {"key": 2}])
        diff = utils.diff_lists_of_objects(list1, list2, {"": "key"})
        self.assertEqual(3, len(diff.added_items()))

    def test_diff_lists_of_objects_only_removals(self):
        list2 = utils.DiffObj("", "", [])
        list1 = utils.DiffObj("", "", [{"key": 1}, {"key": 3}])
        diff = utils.diff_lists_of_objects(list1, list2, {"": "key"})
        self.assertEqual(2, len(diff.removed_items()))

    def test_diff_lists_of_objects_mixed(self):
        list2 = utils.DiffObj("", "", [{"key": 1}, {"key": 3, "a": 1}])
        list1 = utils.DiffObj("", "", [{"key": 1}, {"key": 3, "a": 2, "b": 3}])
        diff = utils.diff_lists_of_objects(list1, list2, {"": "key"})
        self.assertEqual(1, len(diff.modified_items()))

    def test_diff_empty_objects(self):
        obj1 = utils.DiffObj("", "", {})
        diff = utils.diff_objects(obj1, obj1)
        self.assertFalse(diff.is_diff())

    def test_diff_objects_only_additions(self):
        obj1 = utils.DiffObj("", "", {})
        obj2 = utils.DiffObj("", "", {1: 1, 2: 2, 3: 3})
        diff = utils.diff_objects(obj1, obj2)
        self.assertEqual(3, len(diff.added_items()))

    def test_diff_objects_only_removals(self):
        obj2 = utils.DiffObj("", "", {})
        obj1 = utils.DiffObj("", "", {1: 1, 2: 2})
        diff = utils.diff_objects(obj1, obj2)
        self.assertEqual(2, len(diff.removed_items()))

    def test_diff_objects_mixed(self):
        obj1 = utils.DiffObj("", "", {"b": "c", "c": "c", "d": "d"})
        obj2 = utils.DiffObj("", "", {"a": "a", "c": "c", "b": "b"})
        diff = utils.diff_objects(obj1, obj2)
        self.assertEqual(2, len(diff.removed_items()))
        self.assertEqual(2, len(diff.added_items()))

    def test_diff_complex_objects(self):
        obj1 = utils.DiffObj("", "", {"key": 1, "a": [1, 2]})
        obj2 = utils.DiffObj("", "", {"key": 1, "a": [2, 3], "b": {"key": 1}})
        diff = utils.diff_objects(obj1, obj2)
        self.assertEqual(1, len(diff.added_items()))
        self.assertEqual(1, len(diff.modified_items()))

    def test_deep_diff(self):
        obj1 = utils.DiffObj("", "", [
            {"key": 1, "a": [1, 2]},
            {"key": 3, "a": 1, "b": 2},
            {"key": 4},
            {"key": 5, "a": {"data": [{"key": 1}, {"key": 2}]}},
            {
                "key": 6,
                "groups": {"data": [{"name": "group1", "a": "a"}]},
                "mfa": {"data": [{"serial": "abc"}]}
            }])
        obj2 = utils.DiffObj("", "", [
            {"key": 1, "a": [2, 3], "b": {"key": 1}},
            {"key": 3, "a": 2},
            {"key": 2},
            {"key": 5, "a": {"data": [{"key": 3}, {"key": 2}]}},
            {
                "key": 6,
                "groups": {"data": [{"name": "group1", "a": "b"}]},
                "mfa": {"data": []}
            }])
        diff = utils.diff_lists(obj1, obj2, {"": "key", ".b": "key",
                                             ".a.data": "key",
                                             ".groups.data": "name",
                                             ".mfa.data": "serial"})
        self.assertEqual(1, len(diff.added_items()))
        self.assertEqual(1, len(diff.removed_items()))
        self.assertEqual(4, len(diff.modified_items()))
        self.assertEqual(diff.pretty_string(),
                         '[\n\t{\n\t\t\"key\": 1,\n\t\t\"a\": [\n\u001b[31m\t\t\t1\u001b[0m,\n\u001b[32m\t\t\t3\u001b[0m\n\t\t],\n\u001b[32m\t\t\"b\": {\n\t\t\t\"key\": 1\n\t\t}\u001b[0m\n\t},\n\u001b[32m\t{\n\t\t\"key\": 2\n\t}\u001b[0m,\n\t{\n\t\t\"key\": 3,\n\u001b[31m\t\t\"a\": 1\u001b[0m,\n\u001b[32m\t\t\"a\": 2\u001b[0m,\n\u001b[31m\t\t\"b\": 2\u001b[0m\n\t},\n\u001b[31m\t{\n\t\t\"key\": 4\n\t}\u001b[0m,\n\t{\n\t\t\"key\": 5,\n\t\t\"a\": {\n\t\t\t\"data\": [\n\u001b[31m\t\t\t\t{\n\t\t\t\t\t\"key\": 1\n\t\t\t\t}\u001b[0m,\n\u001b[32m\t\t\t\t{\n\t\t\t\t\t\"key\": 3\n\t\t\t\t}\u001b[0m\n\t\t\t]\n\t\t}\n\t},\n\t{\n\t\t\"key\": 6,\n\t\t\"groups\": {\n\t\t\t\"data\": [\n\t\t\t\t{\n\t\t\t\t\t\"name\": \"group1\",\n\u001b[31m\t\t\t\t\t\"a\": \"a\"\u001b[0m,\n\u001b[32m\t\t\t\t\t\"a\": \"b\"\u001b[0m\n\t\t\t\t}\n\t\t\t]\n\t\t},\n\t\t\"mfa\": {\n\t\t\t\"data\": [\n\u001b[31m\t\t\t\t{\n\t\t\t\t\t\"serial\": \"abc\"\n\t\t\t\t}\u001b[0m\n\t\t\t]\n\t\t}\n\t}\n]')  # noqa: E501

    def test_days_elapsed(self):
        now = datetime.utcnow()
        date_time1 = (now - timedelta(days=2)).strftime(utils.DATE_FORMAT)
        print(f"Date time1 {date_time1}")
        time_elapsed = utils.time_diff(date_time1).days
        self.assertEqual(time_elapsed, 2)

        date_time2 = now.strftime(utils.DATE_FORMAT)
        time_elapsed = utils.time_diff(date_time2).days
        self.assertTrue(time_elapsed < 1)

        date_time = (now + timedelta(minutes=2))
        date_time2 = date_time.strftime(utils.DATE_FORMAT)
        time_elapsed = utils.time_diff(date_time2).total_seconds() / 60
        self.assertTrue(time_elapsed < 2)

        date3 = now.strftime("%m/%d/%Y, %H:%M:%S")
        with self.assertRaises(ValueError) as ex:
            utils.time_diff(date3)

        self.assertIn(utils.DATE_FORMAT, str(ex.exception))


if __name__ == '__main__':
    unittest.main()



import json
import unittest
from faros.client import FarosClient
from faros.exceptions import FarosGraphQLException
from graphqlclient import GraphQLClient
from unittest.mock import patch


class FarosClientTest(unittest.TestCase):
    def test_successful_query(self):
        response = {
            "data": {
                "entityThatExists": {
                    "data": []
                }
            }
        }
        event = {"farosToken": "token"}
        query = """{
                  entityThatExists {
                    data {
                      farosParam1
                      farosParam2
                    }
                  }
                }"""

        with patch.object(
                GraphQLClient, "execute", return_value=json.dumps(response)):
            faros_client = FarosClient.from_event(event)
            self.assertEqual(
                faros_client.graphql_query(query), response["data"])

    def test_query_throws_exception(self):
        response = {
            "data": None,
            "errors": [{"message": "some error."}]
        }
        event = {"farosToken": "token"}
        query = """{
                  entityThatDoesNotExist {
                    data {
                      farosParam1
                      farosParam2
                    }
                  }
                }"""

        faros_client = FarosClient.from_event(event)
        with patch.object(
                GraphQLClient, "execute", return_value=json.dumps(response)):
            with self.assertRaises(FarosGraphQLException) as ex:
                faros_client.graphql_query(query)

            self.assertEqual(json.loads(str(ex.exception)), response["errors"])


if __name__ == '__main__':
    unittest.main()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="faros",
    version="0.1.1",
    author="Faros",
    author_email="contact@faros.ai",
    description="Faros API SDK",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/faros-ai/faros-python-client",
    keywords=["faros", "api", "graphql", "sdk"],
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    install_requires=[
        "attrs",
        "graphqlclient",
        "termcolor"
    ],
    python_requires='>=3.6'
)



from __future__ import absolute_import

import attr
import json

from enum import Enum
from datetime import datetime
from termcolor import colored

INDENT = "\t"
DATE_FORMAT = "%Y-%m-%dT%H:%M:%SZ"


@attr.s
class DiffObj:
    """
    Diff object to pass for making comparisons
    """
    name = attr.ib()
    key = attr.ib()
    obj = attr.ib()


@attr.s
class Modification:
    before = attr.ib()
    after = attr.ib()


@attr.s
class Entry:
    key = attr.ib()
    value = attr.ib()

    def __str__(self):
        if self.key:
            return json.dumps(self.key) + ": " + \
                   json.dumps(self.value, indent=INDENT)
        else:
            return json.dumps(self.value, indent=INDENT)


class Modifier(Enum):
    INFO = 0
    REMOVED = 1
    ADDED = 2
    MODIFIED = 3


class DiffType(Enum):
    LIST = 1
    OBJECT = 2


@attr.s
class Diff:
    modifier = attr.ib()
    item = attr.ib()


@attr.s
class JsonDiff:
    """
    A structure representing the differences between two Jsons
    """
    diffs = attr.ib(factory=list)
    key = attr.ib(factory=str)
    diff_type = attr.ib(default=DiffType.LIST)
    object_key = attr.ib(factory=str)

    def prepend(self, item):
        self.diffs.insert(0, item)

    def append(self, item):
        self.diffs.append(item)

    def extend(self, list):
        self.diffs.extend(list)

    def removed_items(self):
        return [x.item for x in self.diffs if x.modifier == Modifier.REMOVED]

    def added_items(self):
        return [x.item for x in self.diffs if x.modifier == Modifier.ADDED]

    def modified_items(self):
        return [x.item for x in self.diffs if x.modifier == Modifier.MODIFIED]

    def is_diff(self):
        return len(self.removed_items()) + len(self.added_items()) + \
               len(self.modified_items()) > 0

    def pretty_string(self, depth=0, outer_type=None):
        def indent(s):
            parts = s.split("\n")
            pad = INDENT * (depth + 1)

            return "\n".join([pad + p for p in parts])

        if not self.diffs:
            return ""

        opener = "[" if self.diff_type == DiffType.LIST else "{"
        closer = "]" if self.diff_type == DiffType.LIST else "}"

        ret = ""
        key = self.key[1:]
        key = key[key.rfind(".") + 1:]
        pad = INDENT * depth
        opener = f"{json.dumps(key)}: {opener}" \
            if key and outer_type == DiffType.OBJECT else opener
        ret += f"{pad}{opener}" if depth else opener
        ret += "\n"
        for diff in self.diffs:
            if diff.modifier == Modifier.REMOVED:
                ret += colored(indent(str(diff.item)), "red")
            elif diff.modifier == Modifier.ADDED:
                ret += colored(indent(str(diff.item)), "green")
            elif diff.modifier == Modifier.INFO:
                ret += indent(str(diff.item))
            else:
                ret += diff.item.pretty_string(depth + 1, self.diff_type)
            ret += ",\n"

        ret = ret[:-2] + "\n"
        ret += f"{pad}{closer}" if depth else closer

        return ret


def diff_lists_of_scalars(list1, list2, prefix=""):
    """
    Compare lists list1 and list2 and returns a delta in color coded format

    :param list list1: list of elements
    :param list list2: list of elements
    :param string prefix: the prefix of the elements in the tree

    :return: the difference in human readable color coded format
    :rtype: list
    """
    list1 = sorted(list1, key=str)
    list2 = sorted(list2, key=str)

    pointer1 = 0
    pointer2 = 0
    diffs = JsonDiff(key=prefix)
    while pointer1 < len(list1) and pointer2 < len(list2):
        item1 = str(list1[pointer1])
        item2 = str(list2[pointer2])

        if item1 < item2:
            diffs.append(Diff(Modifier.REMOVED, item1))
            pointer1 += 1
        elif item1 > item2:
            diffs.append(Diff(Modifier.ADDED, item2))
            pointer2 += 1
        else:
            pointer1 += 1
            pointer2 += 1

    if pointer1 < len(list1):
        diffs.extend([Diff(Modifier.REMOVED, x) for x in list1[pointer1:]])
    if pointer2 < len(list2):
        diffs.extend([Diff(Modifier.ADDED, x) for x in list2[pointer2:]])

    return diffs


def _append_if_modified(outer_diff, inner_diff, key=None, value=None):
    if inner_diff.is_diff():
        if key:
            inner_diff.prepend(Diff(Modifier.INFO, Entry(key, value)))
        outer_diff.append(Diff(Modifier.MODIFIED, inner_diff))


def diff_lists_of_objects(named_list1, named_list2, keys, prefix=""):
    """
    Compares two lists of dicts to find the delta

    :param DiffObj named_list1: first list of objects
    :param DiffObj named_list2: second list of objects
    :param keys: dictionary of keys
    :param prefix: current prefix (padding)
    :return: string representing the differences
    """

    key = keys[prefix]
    list1 = sorted(named_list1.obj, key=lambda x: x[key])
    list2 = sorted(named_list2.obj, key=lambda x: x[key])

    pointer1 = 0
    pointer2 = 0
    diffs = JsonDiff(key=prefix)
    while pointer1 < len(list1) and pointer2 < len(list2):
        item1 = list1[pointer1]
        item2 = list2[pointer2]

        if item1[key] < item2[key]:
            diffs.append(Diff(Modifier.REMOVED, Entry(None, item1)))
            pointer1 += 1
        elif item1[key] > item2[key]:
            diffs.append(Diff(Modifier.ADDED, Entry(None, item2)))
            pointer2 += 1
        else:
            obj1 = DiffObj(named_list1.name, named_list1.key, item1)
            obj2 = DiffObj(named_list2.name, named_list2.key, item2)
            object_diffs = diff_objects(obj1, obj2, keys, prefix)
            _append_if_modified(diffs, object_diffs, key, item1[key])
            pointer1 += 1
            pointer2 += 1

    if pointer1 < len(list1):
        diffs.extend([Diff(Modifier.REMOVED, Entry(None, x))
                      for x in list1[pointer1:]])
    if pointer2 < len(list2):
        diffs.extend([Diff(Modifier.ADDED, Entry(None, x))
                      for x in list2[pointer2:]])

    return diffs


def diff_lists(obj1, obj2, keys={}, prefix=""):
    """
    Compares two lists and returns the delta

    :param DiffObj obj1: list of objects
    :param DiffObj obj2: list of objects
    :param dict keys: key to compare on
    :param str prefix: key prefix

    :return: the delta in the objects
    :rtype: list
    """

    if not obj1.obj and not obj2.obj:
        return JsonDiff(key=prefix)
    elif not obj1.obj and obj2.obj:
        return JsonDiff([Diff(Modifier.ADDED, Entry(None, x))
                         for x in obj2.obj], prefix)
    elif obj1.obj and not obj2.obj:
        return JsonDiff([Diff(Modifier.REMOVED, Entry(None, x))
                         for x in obj1.obj], prefix)
    else:
        if isinstance(obj1.obj[0], dict):
            return diff_lists_of_objects(obj1, obj2, keys, prefix)
        else:
            return diff_lists_of_scalars(obj1.obj, obj2.obj, prefix)


def diff_objects(diff_obj1, diff_obj2, keys=None, prefix=""):  # noqa: C901
    """
    Find the delta between two dict objects. The objects assume to have the
    same types of elements

    :param DiffObj diff_obj1: objects 1 to compare as base
    :param DiffObj diff_obj2: object 2 to compare with
    :param keys: keys to compare in the json
    :param prefix: key prefix when getting object properties

    :return: the delta of the objects
    :rtype list
    """

    if keys is None:
        keys = {}
    obj1 = diff_obj1.obj
    obj2 = diff_obj2.obj

    keys1 = sorted(obj1.keys())
    keys2 = sorted(obj2.keys())

    index1 = 0
    index2 = 0
    diffs = JsonDiff(key=prefix, diff_type=DiffType.OBJECT)
    while index1 < len(keys1) and index2 < len(keys2):
        key1 = keys1[index1]
        key2 = keys2[index2]
        item1 = obj1[key1]
        item2 = obj2[key2]

        if key1 < key2:
            diffs.append(Diff(Modifier.REMOVED, Entry(key1, item1)))
            index1 += 1
        elif key1 > key2:
            diffs.append(Diff(Modifier.ADDED, Entry(key2, item2)))
            index2 += 1
        else:
            named_sub_list1 = DiffObj(diff_obj1.name, diff_obj1.key, item1)
            named_sub_list2 = DiffObj(diff_obj2.name, diff_obj2.key, item2)

            if isinstance(item1, dict):
                dict_diffs = diff_objects(named_sub_list1, named_sub_list2,
                                          keys, f"{prefix}.{key1}")
                _append_if_modified(diffs, dict_diffs)
            elif isinstance(item1, list):
                list_diffs = diff_lists(named_sub_list1, named_sub_list2,
                                        keys, f"{prefix}.{key1}")
                _append_if_modified(diffs, list_diffs)
            elif isinstance(item1, str):
                if item1.replace(diff_obj1.name, "") \
                        != item2.replace(diff_obj2.name, ""):
                    diffs.append(Diff(Modifier.REMOVED, Entry(key1, item1)))
                    diffs.append(Diff(Modifier.ADDED, Entry(key2, item2)))
            else:
                if item1 != item2:
                    diffs.append(Diff(Modifier.REMOVED, Entry(key1, item1)))
                    diffs.append(Diff(Modifier.ADDED, Entry(key2, item2)))

            index1 += 1
            index2 += 1

    if index1 < len(keys1):
        diffs.extend([Diff(Modifier.REMOVED, Entry(x, obj1[x]))
                      for x in keys1[index1:]])
    if index2 < len(keys2):
        diffs.extend([Diff(Modifier.ADDED, Entry(x, obj2[x]))
                      for x in keys2[index2:]])

    return diffs


def time_diff(date_string):
    """
    Gets the time that have elapsed between now and the supplied date string.
    The date string must be of format "%Y-%m-%dT%H:%M:%SZ"

    :param date_string: the date to find the difference between
    :return: number of days that have elapsed
    :rtype: timedelta
    """
    date_diff = datetime.utcnow() - datetime.strptime(date_string, DATE_FORMAT)
    return date_diff
